# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Helpers for Pytest tests with asynchronous programming."""
import asyncio
import logging
import time
from asyncio import AbstractEventLoop
from threading import Thread
from typing import Any, Awaitable, Callable, Optional


def wait_for_condition(
    condition_checker: Callable,
    timeout: int = 2,
    error_msg: str = "Timeout",
    period: float = 0.001,
) -> None:
    """Wait for condition occures in selected timeout."""
    start_time = time.time()

    while not condition_checker():
        time.sleep(period)
        if time.time() > start_time + timeout:
            raise TimeoutError(error_msg)


class AnotherThreadTask:
    """
    Schedule a task to run on the loop in another thread.

    Provides better cancel behaviour: on cancel it will wait till cancelled completely.
    """

    def __init__(self, coro: Awaitable, loop: AbstractEventLoop) -> None:
        """
        Init the task.

        :param coro: coroutine to schedule
        :param loop: an event loop to schedule on.
        """
        self._loop = loop
        self._coro = coro
        self._task: Optional[asyncio.Task] = None
        self._future = asyncio.run_coroutine_threadsafe(self._get_task_result(), loop)

    async def _get_task_result(self) -> Any:
        """
        Get task result, should be run in target loop.

        :return: task result value or raise an exception if task failed
        """
        self._task = self._loop.create_task(self._coro)
        return await self._task

    def result(self, timeout: Optional[float] = None) -> Any:
        """
        Wait for coroutine execution result.

        :param timeout: optional timeout to wait in seconds.
        :return: result
        """
        return self._future.result(timeout)

    def cancel(self) -> None:
        """Cancel coroutine task execution in a target loop."""
        if self._task is None:
            self._loop.call_soon_threadsafe(self._future.cancel)
        else:
            self._loop.call_soon_threadsafe(self._task.cancel)

    def done(self) -> bool:
        """Check task is done."""
        return self._future.done()


class ThreadedAsyncRunner(Thread):
    """Util to run thread with event loop and execute coroutines inside."""

    def __init__(self, loop: Optional[AbstractEventLoop] = None) -> None:
        """
        Init threaded runner.

        :param loop: optional event loop. is it's running loop, threaded runner will use it.
        """
        self._loop = loop or asyncio.new_event_loop()
        if self._loop.is_closed():
            raise ValueError("Event loop closed.")  # pragma: nocover
        super().__init__(daemon=True)

    def start(self) -> None:
        """Start event loop in dedicated thread."""
        if self.is_alive() or self._loop.is_running():  # pragma: nocover
            return
        super().start()
        self.call(asyncio.sleep(0.001)).result(1)

    def run(self) -> None:
        """Run code inside thread."""
        logging.debug("Starting threaded asyncio loop...")
        asyncio.set_event_loop(self._loop)
        self._loop.run_forever()
        logging.debug("Asyncio loop has been stopped.")

    def call(self, coro: Awaitable) -> Any:
        """
        Run a coroutine inside the event loop.

        :param coro: a coroutine to run.
        :return: task
        """
        return AnotherThreadTask(coro, self._loop)

    def stop(self) -> None:
        """Stop event loop in thread."""
        logging.debug("Stopping...")

        if not self.is_alive():  # pragma: nocover
            return

        if self._loop.is_running():
            logging.debug("Stopping loop...")
            self._loop.call_soon_threadsafe(self._loop.stop)

        logging.debug("Wait thread to join...")
        self.join(10)
        logging.debug("Stopped.")


class BaseThreadedAsyncLoop:
    """Test class with a threaded event loop running."""

    DEFAULT_ASYNC_TIMEOUT = 5.0
    loop: ThreadedAsyncRunner
    thread: Thread

    def setup(self) -> None:
        """Set up the class."""
        self.loop = ThreadedAsyncRunner()
        self.loop.start()

    def execute(self, coro: Awaitable, timeout: float = DEFAULT_ASYNC_TIMEOUT) -> Any:
        """Execute a coroutine and wait its completion."""
        task: AnotherThreadTask = self.loop.call(coro)
        return task.result(timeout=timeout)

    def teardown(self) -> None:
        """Teardown the class."""
        self.loop.start()
        self.loop.join(5.0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Helpers for Tendermint."""
import logging
import subprocess  # nosec
from pathlib import Path
from typing import Any, List

import pytest
from aea_test_autonomy.configurations import LOCALHOST
from aea_test_autonomy.helpers.base import tendermint_health_check


_TCP = "tcp://"
_HTTP = "http://"
_LOCAL_ADDRESS = "0.0.0.0"  # nosec
_STARTING_ABCI_PORT = 26658
_STARTING_RPC_PORT = 26657
_STARTING_P2P_PORT = 26656


class TendermintNodeInfo:
    """Data class to store Tendermint node info."""

    def __init__(
        self, node_id: str, abci_port: int, rpc_port: int, p2p_port: int, home: Path
    ):
        """Initialize Tendermint node info."""
        self.node_id = node_id
        self.abci_port = abci_port
        self.rpc_port = rpc_port
        self.p2p_port = p2p_port
        self.home = home

    @property
    def rpc_laddr(self) -> str:
        """Get ith rpc_laddr."""
        return f"{_TCP}{_LOCAL_ADDRESS}:{self.rpc_port}"

    def get_http_addr(self, host: str) -> str:
        """Get ith HTTP RCP address, given the host."""
        return f"{_HTTP}{host}:{self.rpc_port}"

    @property
    def p2p_laddr(self) -> str:
        """Get ith p2p_laddr."""
        return f"{_TCP}{_LOCAL_ADDRESS}:{self.p2p_port}"


class TendermintLocalNetworkBuilder:
    """Build a local Tendermint network."""

    def __init__(
        self,
        nb_nodes: int,
        directory: Path,
        consensus_create_empty_blocks: bool = True,
    ) -> None:
        """Initialize the builder."""
        self.nb_nodes = nb_nodes
        self.directory = Path(directory).resolve().absolute()
        self.consensus_create_empty_blocks = consensus_create_empty_blocks

        self._create()

    def _create(self) -> None:
        """Create a Tendermint local network."""
        self._create_testnet()
        self.nodes = [
            TendermintNodeInfo(
                self._get_node_id(i),
                _STARTING_ABCI_PORT + i * 10,
                _STARTING_RPC_PORT + i * 10,
                _STARTING_P2P_PORT + i * 10,
                self.directory / f"node{i}",
            )
            for i in range(self.nb_nodes)
        ]

    def _create_testnet(self) -> None:
        """Create a testnet calling 'tendermint testnet'."""
        subprocess.call(  # nosec
            [
                "tendermint",
                "testnet",
                "--v",
                str(self.nb_nodes),
                "--o",
                str(self.directory),
            ]
        )

    def _get_node_id(self, i: int) -> str:  # pylint: disable=no-self-use
        """Get the node id."""
        assert 0 <= i < self.nb_nodes, "invalid node id"  # nosec
        node_name = f"node{i}"
        process = subprocess.Popen(  # nosec  # pylint: disable=consider-using-with
            ["tendermint", "--home", node_name, "show-node-id"], stdout=subprocess.PIPE
        )
        output, _ = process.communicate()
        node_id = output.decode().strip()
        return node_id

    def get_p2p_seeds(self) -> List[str]:
        """Get p2p seeds."""
        return [f"{_TCP}{n.node_id}@{_LOCAL_ADDRESS}:{n.p2p_port}" for n in self.nodes]

    def get_command(self, i: int) -> List[str]:
        """Get command-line command for the ith process."""
        n = self.nodes[i]
        return [
            "tendermint",
            "node",
            "--home",
            str(n.home),
            f"--rpc.laddr={n.rpc_laddr}",
            f"--p2p.laddr={n.p2p_laddr}",
            f"--p2p.seeds={','.join(self.get_p2p_seeds())}",
            f"--consensus.create_empty_blocks={self.consensus_create_empty_blocks}",
        ]

    @property
    def http_rpc_laddrs(self) -> List[str]:
        """Get HTTP RPC listening addresses."""
        return [node.get_http_addr(LOCALHOST) for node in self.nodes]


class BaseTendermintTestClass:  # pylint: disable=too-few-public-methods
    """MixIn class for Pytest classes."""

    @staticmethod
    def health_check(
        tendermint_net: TendermintLocalNetworkBuilder, **kwargs: Any
    ) -> None:
        """Do a health-check of the Tendermint network."""
        logging.info("Waiting for Tendermint nodes to be up")
        for rpc_addr in tendermint_net.http_rpc_laddrs:
            if not tendermint_health_check(rpc_addr, **kwargs):
                pytest.fail(f"Tendermint node {rpc_addr} did not pass health-check")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains testing utilities."""
import logging
import platform
import re
import shutil
import subprocess  # nosec
import time
from abc import ABC, abstractmethod
from typing import Any, Dict, Generator, List

import docker
import pytest
from docker.errors import ImageNotFound, NotFound
from docker.models.containers import Container


SEPARATOR = ("\n" + "*" * 40) * 3 + "\n"
logger = logging.getLogger(__name__)


skip_docker_tests = pytest.mark.skipif(
    platform.system() != "Linux",
    reason="Docker daemon is not available in Windows and macOS CI containers.",
)


class DockerImage(ABC):
    """A class to wrap interaction with a Docker image."""

    MINIMUM_DOCKER_VERSION = (19, 0, 0)

    def __init__(self, client: docker.DockerClient):
        """Initialize."""
        self._client = client

    def check_skip(self) -> None:
        """
        Check whether the test should be skipped.

        By default, nothing happens.
        """
        self._check_docker_binary_available()

    def _check_docker_binary_available(self) -> None:
        """Check the 'Docker' CLI tool is in the OS PATH."""
        result = shutil.which("docker")
        if result is None:
            pytest.skip("Docker not in the OS Path; skipping the test")

        result_ = subprocess.run(  # nosec   # pylint: disable=subprocess-run-check
            ["docker", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if result_.returncode != 0:
            pytest.skip(
                f"'docker --version' failed with exit code {result_.returncode}"
            )

        match = re.search(
            r"Docker version ([0-9]+)\.([0-9]+)\.([0-9]+)",
            result_.stdout.decode("utf-8"),
        )
        if match is None:
            pytest.skip("cannot read version from the output of 'docker --version'")
            return
        version = (int(match.group(1)), int(match.group(2)), int(match.group(3)))
        if version < self.MINIMUM_DOCKER_VERSION:
            pytest.skip(
                f"expected Docker version to be at least {'.'.join(*[str(el) for el in self.MINIMUM_DOCKER_VERSION])}, found {'.'.join(*[str(el) for el in version])}"
            )

    @property
    @abstractmethod
    def image(self) -> str:
        """Return the image name."""

    def stop_if_already_running(self) -> None:
        """Stop the running images with the same tag, if any."""
        client = docker.from_env()
        for container in client.containers.list():
            if self.image in container.image.tags:
                logger.info(f"Stopping image {self.image}...")
                container.stop()

    @abstractmethod
    def create(self) -> Container:
        """Instantiate the image in a container."""

    @abstractmethod
    def create_many(self, nb_containers: int) -> List[Container]:
        """Instantiate the image in many containers, parametrized."""

    @abstractmethod
    def wait(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        """
        Wait until the image is running.

        :param max_attempts: max number of attempts.
        :param sleep_rate: the amount of time to sleep between different requests.
        :return: True if the wait was successful, False otherwise.
        """


def _pre_launch(image: DockerImage) -> None:
    """Run pre-launch checks."""
    image.check_skip()
    image.stop_if_already_running()


def _start_container(
    image: DockerImage, container: Container, timeout: float, max_attempts: int
) -> None:
    """
    Start a container.

    :param image: an instance of Docker image.
    :param container: the container to start, created from the image.
    :param timeout: timeout to launch
    :param max_attempts: max launch attempts
    """
    container.start()
    logger.info(f"Setting up image {image.image}...")
    success = image.wait(max_attempts, timeout)
    if not success:
        container.stop()
        logger.error(
            f"{SEPARATOR}Logs from container {container.name}:\n{container.logs().decode()}"
        )
        container.remove()
        pytest.fail(f"{image.image} doesn't work. Exiting...")
    else:
        logger.info("Done!")
        time.sleep(timeout)


def _stop_container(container: Container, tag: str) -> None:
    """Stop a container."""
    logger.info(f"Stopping container {container.name} from image {tag}...")
    container.stop()
    try:
        logger.info(
            f"{SEPARATOR}Logs from container {container.name}:\n{container.logs().decode()}"
        )
        if str(container.name).startswith("node"):
            logger.info(f"{SEPARATOR}Logs from container log file {container.name}:\n")
            bits, _ = container.get_archive(f"/logs/{container.name}.txt")
            for chunk in bits:
                logger.info(chunk.decode())
    except (ImageNotFound, NotFound) as e:
        logger.error(e)
    finally:
        container.remove()


def launch_image(
    image: DockerImage, timeout: float = 2.0, max_attempts: int = 10
) -> Generator[DockerImage, None, None]:
    """
    Launch a single container.

    :param image: an instance of Docker image.
    :param timeout: timeout to launch
    :param max_attempts: max launch attempts
    :yield: image
    """
    _pre_launch(image)
    container = image.create()
    _start_container(image, container, timeout, max_attempts)
    yield image
    _stop_container(container, image.image)


def launch_many_containers(
    image: DockerImage, nb_containers: int, timeout: float = 2.0, max_attempts: int = 10
) -> Generator[DockerImage, None, None]:
    """
    Launch many containers from an image.

    :param image: an instance of Docker image.
    :param nb_containers: the number of containers to launch from the image.
    :param timeout: timeout to launch
    :param max_attempts: max launch attempts
    :yield: image
    """
    _pre_launch(image)
    containers = image.create_many(nb_containers)
    for container in containers:
        _start_container(image, container, timeout, max_attempts)
    yield image
    for container in containers:
        _stop_container(container, image.image)


class DockerBaseTest(ABC):
    """Base pytest class for setting up Docker images."""

    timeout: float = 3.0
    max_attempts: int = 60
    addr: str
    port: int

    _image: DockerImage
    _container: Container

    @classmethod
    def setup_class(cls) -> None:
        """Setup up the test class."""
        cls._image = cls._build_image()
        cls._image.check_skip()
        cls._image.stop_if_already_running()
        cls._container = cls._image.create()
        cls._container.start()
        logger.info(f"Setting up image {cls._image.image}...")
        success = cls._image.wait(cls.max_attempts, cls.timeout)
        if not success:
            cls._container.stop()
            logger.error(
                f"{SEPARATOR}Logs from container {cls._container.name}:\n{cls._container.logs().decode()}"
            )
            cls._container.remove()
            pytest.fail(f"{cls._image.image} doesn't work. Exiting...")
        else:
            logger.info("Done!")
            time.sleep(cls.timeout)
        cls._setup_class(**cls.setup_class_kwargs())

    @classmethod
    def teardown_class(cls) -> None:
        """Tear down the test."""
        logger.info(f"Stopping the image {cls._image.image}...")
        cls._container.stop()
        logger.info(
            f"{SEPARATOR}Logs from container {cls._container.name}:\n{cls._container.logs().decode()}"
        )
        cls._container.remove()

    @classmethod
    @abstractmethod
    def _build_image(cls) -> DockerImage:
        """Instantiate the Docker image."""

    @classmethod
    @abstractmethod
    def _setup_class(cls, **setup_class_kwargs: Any) -> None:
        """Continue setting up the class."""

    @classmethod
    @abstractmethod
    def setup_class_kwargs(cls) -> Dict[str, Any]:
        """Get kwargs for _setup_class call."""



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Base test classes."""

import os
import time
from abc import ABC
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Optional, Tuple, cast

from aea.contracts.base import Contract
from aea.crypto.base import Crypto, LedgerApi
from aea.crypto.registries import crypto_registry, ledger_apis_registry
from aea_ledger_ethereum import (
    DEFAULT_CHAIN_ID,
    DEFAULT_CURRENCY_DENOM,
    DEFAULT_EIP1559_STRATEGY,
    DEFAULT_GAS_STATION_STRATEGY,
    EthereumCrypto,
)
from aea_test_autonomy.fixture_helpers import (
    GanacheBaseTest,
    HardHatAMMBaseTest,
    HardHatGnosisBaseTest,
)
from aea_test_autonomy.helpers.contracts import get_register_contract


class BaseContractTest(ABC):
    """Base test class for contract tests."""

    contract_directory: Path
    contract: Contract
    ledger_api: LedgerApi
    identifier: str = EthereumCrypto.identifier
    deployer_crypto: Crypto
    contract_address: Optional[str] = None

    @classmethod
    def _setup_class(cls, **kwargs: Any) -> None:
        """Setup test."""
        key_pairs: List[Tuple[str, str]] = kwargs.pop("key_pairs")
        url: str = kwargs.pop("url")
        new_config = {
            "address": url,
            "chain_id": DEFAULT_CHAIN_ID,
            "denom": DEFAULT_CURRENCY_DENOM,
            "default_gas_price_strategy": "eip1559",
            "gas_price_strategies": {
                "gas_station": DEFAULT_GAS_STATION_STRATEGY,
                "eip1559": DEFAULT_EIP1559_STRATEGY,
            },
        }
        cls.contract = get_register_contract(cls.contract_directory)
        cls.ledger_api = ledger_apis_registry.make(
            cls.identifier,
            **new_config,
        )
        with TemporaryDirectory() as temp_dir:
            output_file = Path(os.path.join(temp_dir, "key_file"))
            with open(  # pylint: disable=unspecified-encoding
                output_file, "w"
            ) as text_file:
                text_file.write(key_pairs[0][1])
            cls.deployer_crypto = crypto_registry.make(
                cls.identifier, private_key_path=output_file
            )
        cls.deploy(**cls.deployment_kwargs())
        assert cls.contract_address is not None, "Contract not deployed."  # nosec

    @classmethod
    def deploy(cls, **kwargs: Any) -> None:
        """Deploy the contract."""
        tx = cls.contract.get_deploy_transaction(
            ledger_api=cls.ledger_api,
            deployer_address=str(cls.deployer_crypto.address),
            **kwargs,
        )
        if tx is None:
            return None
        contract_address = tx.pop("contract_address", None)
        tx_signed = cls.deployer_crypto.sign_transaction(tx)
        tx_hash = cls.ledger_api.send_signed_transaction(tx_signed)
        if tx_hash is None:
            return None
        time.sleep(0.5)  # give it time to mine the block
        tx_receipt = cls.ledger_api.get_transaction_receipt(tx_hash)
        if tx_receipt is None:
            return None
        contract_address = (
            cast(Dict, tx_receipt)["contractAddress"]
            if contract_address is None
            else contract_address
        )
        cls.contract_address = cast(str, contract_address)
        return None

    @classmethod
    def deployment_kwargs(cls) -> Dict[str, Any]:
        """Get deployment kwargs."""
        raise NotImplementedError


class BaseGanacheContractTest(BaseContractTest, GanacheBaseTest):
    """Base test case for testing contracts on Ganache."""


class BaseHardhatGnosisContractTest(BaseContractTest, HardHatGnosisBaseTest):
    """Base test case for testing contracts on Hardhat with Gnosis."""


class BaseHardhatAMMContractTest(BaseContractTest, HardHatAMMBaseTest):
    """Base test case for testing AMM contracts on Hardhat."""


class BaseContractWithDependencyTest(BaseContractTest):
    """Base test contract with contract dependencies"""

    dependencies: List[Tuple[str, Path, Any]] = []
    dependency_info: Dict[str, Tuple[str, Contract]] = {}

    @classmethod
    def _deploy_dependencies(cls) -> None:
        """Deploy the dependencies"""

        original_contract_directory = cls.contract_directory
        original_contract = cls.contract

        for dependency in cls.dependencies:
            label, path, kwargs = dependency
            cls.contract_directory = path
            cls.contract = get_register_contract(cls.contract_directory)

            deps = kwargs.pop("deps", {})  # dependencies this contract has

            for dep_name, dep in deps.items():
                dep_address, _ = cls.dependency_info[dep]
                kwargs[dep_name] = dep_address

            cls.deploy(**kwargs)

            cls.dependency_info[label] = (str(cls.contract_address), cls.contract)

        # reset class vars to their initial state
        cls.contract_directory = original_contract_directory
        cls.contract = original_contract
        cls.contract_address = None

    @classmethod
    def _setup_class(cls, **kwargs: Any) -> None:
        """Deploy the dependencies then setups the contract"""
        key_pairs: List[Tuple[str, str]] = kwargs.pop("key_pairs")
        url: str = kwargs.pop("url")
        new_config = {
            "address": url,
            "chain_id": DEFAULT_CHAIN_ID,
            "denom": DEFAULT_CURRENCY_DENOM,
            "default_gas_price_strategy": "eip1559",
            "gas_price_strategies": {
                "gas_station": DEFAULT_GAS_STATION_STRATEGY,
                "eip1559": DEFAULT_EIP1559_STRATEGY,
            },
        }
        cls.contract = get_register_contract(cls.contract_directory)
        cls.ledger_api = ledger_apis_registry.make(
            cls.identifier,
            **new_config,
        )
        with TemporaryDirectory() as temp_dir:
            output_file = Path(os.path.join(temp_dir, "key_file"))
            with open(  # pylint: disable=unspecified-encoding
                output_file, "w"
            ) as text_file:
                text_file.write(key_pairs[0][1])
            cls.deployer_crypto = crypto_registry.make(
                cls.identifier, private_key_path=output_file
            )

        cls._deploy_dependencies()

        cls.deploy(**cls.deployment_kwargs())
        assert cls.contract_address is not None, "Contract not deployed."  # nosec


class BaseGanacheContractWithDependencyTest(
    BaseContractWithDependencyTest, GanacheBaseTest
):
    """Base test case for testing contracts on Ganache."""


class BaseHardhatGnosisContractWithDependencyTest(
    BaseContractWithDependencyTest, HardHatGnosisBaseTest
):
    """Base test case for testing contracts on Hardhat with Gnosis."""


class BaseHardhatAMMContractWithDependencyTest(
    BaseContractWithDependencyTest, HardHatAMMBaseTest
):
    """Base test case for testing AMM contracts on Hardhat."""



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tendermint Docker image."""
import os
import re
import subprocess  # nosec
import time
from pathlib import Path
from typing import Any, Dict, List

import docker
import pytest
from aea_test_autonomy.configurations import (
    TENDERMINT_IMAGE_NAME,
    TENDERMINT_IMAGE_VERSION,
)
from aea_test_autonomy.docker.base import DockerImage
from aea_test_autonomy.helpers.base import tendermint_health_check
from docker.models.containers import Container


_TCP = "tcp://"
_HTTP = "http://"
_LOCAL_ADDRESS = "0.0.0.0"  # nosec

DEFAULT_TENDERMINT_PORT = 26657
DEFAULT_P2P_PORT = 26656
DEFAULT_TENDERMINT_COM_PORT = 8080
DEFAULT_ABCI_PORT = 26658
# we need this because we want to connect from the Tendermint
# Docker container to the ABCI server that lives in the host
DEFAULT_ABCI_HOST = "host.docker.internal"

_SLEEP_TIME = 1


class TendermintDockerImage(DockerImage):
    """Tendermint Docker image."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        client: docker.DockerClient,
        abci_host: str = DEFAULT_ABCI_HOST,
        abci_port: int = DEFAULT_ABCI_PORT,
        port: int = DEFAULT_TENDERMINT_PORT,
        p2p_port: int = DEFAULT_P2P_PORT,
        com_port: int = DEFAULT_TENDERMINT_COM_PORT,
    ):
        """Initialize."""
        super().__init__(client)
        self.abci_host = abci_host
        self.abci_port = abci_port
        self.port = port
        self.p2p_port = p2p_port
        self.com_port = com_port
        self.proxy_app = f"{_TCP}{self.abci_host}:{self.abci_port}"

    @property
    def image(self) -> str:
        """Get the image name."""
        return "tendermint/tendermint:v0.34.19"

    def _build_command(self) -> List[str]:
        """Build command."""
        cmd = ["node", f"--proxy_app={self.proxy_app}"]
        return cmd

    def create(self) -> Container:
        """Create the container."""
        cmd = self._build_command()
        ports = {
            f"{DEFAULT_TENDERMINT_PORT}/tcp": (_LOCAL_ADDRESS, self.port),  # nosec
        }
        if self.abci_host == DEFAULT_ABCI_HOST:
            extra_hosts_config = {self.abci_host: "host-gateway"}
        else:
            extra_hosts_config = {}
        container = self._client.containers.run(
            self.image,
            command=cmd,
            detach=True,
            ports=ports,
            extra_hosts=extra_hosts_config,
        )
        return container

    def create_many(self, nb_containers: int) -> List[Container]:
        """Instantiate the image in many containers, parametrized."""
        raise NotImplementedError()

    def wait(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        """
        Wait until the image is running.

        :param max_attempts: max number of attempts.
        :param sleep_rate: the amount of time to sleep between different requests.
        :return: True if the wait was successful, False otherwise.
        """
        time.sleep(_SLEEP_TIME)
        return True


class FlaskTendermintDockerImage(TendermintDockerImage):
    """Flask app with Tendermint Docker image."""

    _extra_hosts: Dict[str, str]

    def __init__(  # pylint: disable=too-many-arguments,useless-super-delegation
        self,
        client: docker.DockerClient,
        abci_host: str = DEFAULT_ABCI_HOST,
        abci_port: int = DEFAULT_ABCI_PORT,
        port: int = DEFAULT_TENDERMINT_PORT,
        p2p_port: int = DEFAULT_P2P_PORT,
        com_port: int = DEFAULT_TENDERMINT_COM_PORT + 2,
    ):
        """Initialize."""
        super().__init__(client, abci_host, abci_port, port, p2p_port, com_port)

    @property
    def image(self) -> str:
        """Get the image name."""
        return f"{TENDERMINT_IMAGE_NAME}:{TENDERMINT_IMAGE_VERSION}"

    @staticmethod
    def get_node_name(i: int) -> str:
        """Get the ith node's name."""
        return f"node{i}"

    def _get_node_id(self, i: int) -> str:
        """Get the node id."""
        cmd = [
            "docker",
            "exec",
            self.get_node_name(i),
            "tendermint",
            "--home",
            self.get_node_name(i),
            "show-node-id",
        ]
        process = subprocess.Popen(  # nosec    # pylint: disable=consider-using-with
            cmd,
            stdout=subprocess.PIPE,
        )
        output, _ = process.communicate()
        node_id = output.decode().strip()
        return node_id

    @staticmethod
    def __increment_port(port: int, i: int) -> int:
        """Increment a port"""
        return port + i * 10

    def get_port(self, i: int) -> int:
        """Get the ith port."""
        return self.__increment_port(self.port, i)

    def get_com_port(self, i: int) -> int:
        """Get the ith com port."""
        return self.__increment_port(self.com_port, i)

    def get_p2p_port(self, i: int) -> int:
        """Get the ith p2p port."""
        return self.__increment_port(self.p2p_port, i)

    def get_abci_port(self, i: int) -> int:
        """Get the ith abci port."""
        return self.__increment_port(self.abci_port, i)

    def get_addr(self, prefix: str, i: int, p2p: bool = False) -> str:
        """Get a node's address."""
        valid_prefixes = {_TCP, _HTTP}
        if prefix not in valid_prefixes:
            raise ValueError(f"Invalid prefix! Should be one of: {valid_prefixes}")

        if p2p:
            prefix += self._get_node_id(i) + "@"
            port = self.get_p2p_port(i)
        else:
            port = self.get_port(i)

        return f"{prefix}{_LOCAL_ADDRESS}:{port}"

    @property
    def p2p_seeds(self) -> List[str]:
        """Get p2p seeds."""
        if self.nb_nodes is None:
            raise ValueError("Trying to get p2p seeds before initializing containers!")

        return [self.get_addr(_TCP, i) for i in range(self.nb_nodes)]

    def _build_command(self) -> List[str]:
        """Build command."""
        return ["run", "--no-reload", f"--host={_LOCAL_ADDRESS}", "--port=8080"]

    def _create_one(self, i: int) -> Container:
        """Create a node container."""
        name = self.get_node_name(i)
        extra_hosts = (
            {self.abci_host: "host-gateway"}
            if self.abci_host == DEFAULT_ABCI_HOST
            else {}
        )
        extra_hosts.update(self._extra_hosts)

        run_kwargs = dict(
            image=self.image,
            command=self._build_command(),
            name=name,
            hostname=name,
            detach=True,
            mem_limit="1024m",
            mem_reservation="256M",
            environment={
                "ID": i,
                "PROXY_APP": f"{_TCP}{self.abci_host}:{self.get_abci_port(i)}",
                "TMHOME": f"/tendermint/{name}",
                "CREATE_EMPTY_BLOCKS": "true",
                "DEV_MODE": "1",
                "LOG_FILE": f"/logs/{name}.txt",
            },
            working_dir="/tendermint",
            volumes=[
                f"{os.getcwd()}/nodes:/tendermint:Z",
                f"{os.getcwd()}/logs:/logs:Z",
                f"{os.getcwd()}/tm_state:/tm_state:Z",
            ],
            ports={
                f"{DEFAULT_TENDERMINT_PORT}/tcp": (
                    _LOCAL_ADDRESS,  # nosec,
                    self.get_port(i),
                ),
                f"{DEFAULT_TENDERMINT_COM_PORT}/tcp": (
                    _LOCAL_ADDRESS,  # nosec,
                    self.get_com_port(i),
                ),
                f"{DEFAULT_P2P_PORT}/tcp": (
                    _LOCAL_ADDRESS,  # nosec,
                    self.get_p2p_port(i),
                ),
            },
            extra_hosts=extra_hosts,
        )
        container = self._client.containers.run(**run_kwargs)
        return container

    def _fix_persistent_peers(self) -> None:  # pylint: disable=too-many-locals
        """
        Fix the persistent peers' ports in the configuration file.

        Since we are running all the ABCIs at the same host for our e2e tests, we shift the ports by 10 for each
        added node. Therefore, we need to override the default persistent peers in the config files,
        in order for them to use the correct ports.
        """
        nodes_config_files = list(Path().cwd().glob("**/config.toml"))
        assert (  # nosec
            nodes_config_files != []
        ), "Could not detect any config files for the nodes!"

        for config_file in nodes_config_files:
            config_text = config_file.read_text(encoding="utf-8")
            peers = re.findall(r"[a-z\d]+@node\d:\d+", config_text)

            updated_peers = []
            for peer in peers:
                peer_id, address = peer.split("@")
                peer_name, _ = address.split(":")
                *_, peer_number_string = peer_name

                peer_number = int(peer_number_string)
                new_port = self.get_p2p_port(peer_number)
                updated_peers.append(f"{peer_id}@{peer_name}:{new_port}")

            persistent_peers_string = (
                'persistent_peers = "' + ",".join(updated_peers) + '"\n'
            )
            updated_config = re.sub(
                'persistent_peers = ".*\n', persistent_peers_string, config_text
            )

            config_file.write_text(updated_config, encoding="utf-8")

    def _grant_permissions(self) -> None:
        """
        Grant permissions for the nodes' config files.

        Create the nodes' config files, so that the `testnet` command which is run via docker
        does not create it with root permissions, so we can later modify the `config.toml` files.
        """
        for i in range(self.nb_nodes):
            path = Path(f"{os.getcwd()}", "nodes", f"node{i}", "config")
            os.makedirs(path)
            open(  # pylint: disable=consider-using-with,unspecified-encoding
                path / "config.toml", "a"
            ).close()

    def _create_testnet(self) -> None:
        """Create the Tendermint testnet."""
        cmd = [
            "docker",
            "run",
            "--rm",
            "-v",
            f"{os.getcwd()}/nodes:/tendermint:Z",
            "--entrypoint=/usr/bin/tendermint",
            self.image,
            "testnet",
            "--config",
            "/etc/tendermint/config-template.toml",
            "--v",
            f"{self.nb_nodes}",
            "--o",
            "/tendermint/",
        ]
        for i in range(self.nb_nodes):
            cmd.append(f"--hostname=node{i}")

        subprocess.run(cmd)  # nosec  # pylint: disable=subprocess-run-check

    def _create_config(self, nb_nodes: int) -> None:
        """Create necessary configuration."""
        self.nb_nodes = nb_nodes  # pylint: disable=attribute-defined-outside-init
        self._grant_permissions()
        self._create_testnet()
        self._fix_persistent_peers()
        self._extra_hosts = {
            self.get_node_name(i): "host-gateway" for i in range(nb_nodes)
        }

    def create_many(self, nb_containers: int) -> List[Container]:
        """Create a list of node containers."""
        self._create_config(nb_containers)
        containers = [self._create_one(i) for i in range(nb_containers)]
        return containers

    def health_check(self, **kwargs: Any) -> None:
        """Do a health-check of the Tendermint network."""
        http_rpc_laddresses = [self.get_addr(_HTTP, i) for i in range(self.nb_nodes)]
        for http_rpc_laddr in http_rpc_laddresses:
            if not tendermint_health_check(http_rpc_laddr, **kwargs):
                pytest.fail(
                    f"Tendermint node {http_rpc_laddr} did not pass health-check"
                )
